# init file
"""Package initialisation file"""

__author__  = "Mark Slater <mws@hep.ph.bham.ac.uk>"
__date__    = "11 February 2009"
__version__ = "1.0"

from Spider import Spider
from SpiderRTHandler import *
